from sqlalchemy import ForeignKey, JSON
from sqlalchemy.orm import Mapped, mapped_column, relationship

from artemis_model.base import TimeStampMixin, CustomSyncBase, CustomBase

from sqlalchemy.ext.declarative import declared_attr


class SettingMixin(TimeStampMixin):
    """
    Exclusion Schema
    {
        "genres": {
            "<genre_id>": "<genre_name>",
            "<genre_id>": "<genre_name>",
            ...
        },
        "artists": {
            "<artist_id>": "<artist_name>",
            "<artist_id>": "<artist_name>",
            ...
        },
        "tracks": {
            "<track_id>": "<track_name>",
            "<track_id>": "<track_name>",
            ...
        },
        "exclude_pal": True/False
    }
    """

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    zone_id: Mapped[int] = mapped_column(ForeignKey("zone.id"), index=True)
    exclusion: Mapped[JSON] = mapped_column(JSON, nullable=True)

    @declared_attr
    def zone(cls) -> Mapped["Zone"]:
        return relationship(back_populates="setting")


class SettingSync(CustomSyncBase, SettingMixin):
    pass


class Setting(CustomBase, SettingMixin):
    pass



"""Redis models."""

from .zone_state import ZoneState, NowPlaying, SessionId, BucketId
from .device import ActiveDevice
from .bucket import RedisTrackBucketItem
from .play_history import RedisTrackPlayHistoryItem
from .keys import (
    KEY_ZONE_PLAY_HISTORY_TEMPLATE,
    KEY_ZONE_PUSH_PLAYLIST_BUCKET_TEMPLATE,
    KEY_ZONE_SCHEDULE_BUCKET_TEMPLATE,
    KEY_ZONE_STATE_TEMPLATE,
    KEY_ZONE_ACTIVE_DEVICE_TEMPLATE,
)


__all__ = [
    "ZoneState",
    "NowPlaying",
    "SessionId",
    "BucketId",
    "ActiveDevice",
    "KEY_ZONE_PLAY_HISTORY_TEMPLATE",
    "KEY_ZONE_PUSH_PLAYLIST_BUCKET_TEMPLATE",
    "KEY_ZONE_SCHEDULE_BUCKET_TEMPLATE",
    "KEY_ZONE_STATE_TEMPLATE",
    "KEY_ZONE_ACTIVE_DEVICE_TEMPLATE",
    "RedisTrackBucketItem",
    "RedisTrackPlayHistoryItem",
]



import uuid
from typing import List
from sqlalchemy import JSON, ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship

from artemis_model.base import CustomSyncBase, TimeStampMixin, AuditMixin, CustomBase

from sqlalchemy.ext.declarative import declared_attr


class MessageMixin(TimeStampMixin, AuditMixin):
    """
    This table is used to store the message.
    """

    id: Mapped[int] = mapped_column(autoincrement=True, primary_key=True, index=True)
    name: Mapped[str] = mapped_column(nullable=False)
    disabled: Mapped[bool] = mapped_column(nullable=False, default=False)
    s3_link: Mapped[str] = mapped_column(nullable=False)

    organization_id: Mapped[uuid.UUID] = mapped_column(
        ForeignKey("organization.id"), nullable=False, index=True
    )
    message_group_id: Mapped[int] = mapped_column(
        ForeignKey("message_group.id"), nullable=True, index=True, default=None
    )

    @declared_attr
    def organization(cls) -> Mapped["Organization"]:
        return relationship("Organization", back_populates="messages")

    @declared_attr
    def message_group(cls) -> Mapped["MessageGroup"]:
        return relationship("MessageGroup", back_populates="messages")

    @declared_attr
    def play_details(cls) -> Mapped[List["MessagePlayDetail"]]:
        return relationship("MessagePlayDetail", back_populates="message")


class MessageSync(CustomSyncBase, MessageMixin):
    pass


class Message(CustomBase, MessageMixin):
    pass


class MessagePlayDetailMixin(TimeStampMixin, AuditMixin):
    """
    This table is used to store the message play details.
    Frequency is used only for rotated messages.
    Timesheet is used only for scheduled messages.
    """

    id: Mapped[int] = mapped_column(autoincrement=True, primary_key=True, index=True)
    message_id: Mapped[int] = mapped_column(ForeignKey("message.id"), nullable=False, index=True)
    zone_id: Mapped[int] = mapped_column(ForeignKey("zone.id"), index=True)

    type: Mapped[str] = mapped_column(nullable=False)

    timesheet: Mapped[JSON] = mapped_column(JSON, nullable=True)

    @declared_attr
    def message(cls) -> Mapped["Message"]:
        return relationship("Message", back_populates="play_details")

    @declared_attr
    def zone(cls) -> Mapped["Zone"]:
        return relationship("Zone", back_populates="message_play_details")


class MessagePlayDetailSync(CustomSyncBase, MessagePlayDetailMixin):
    pass


class MessagePlayDetail(CustomBase, MessagePlayDetailMixin):
    pass


class MessageGroupMixin(TimeStampMixin, AuditMixin):
    """
    This table is used to store the message group.
    """

    id: Mapped[int] = mapped_column(autoincrement=True, primary_key=True, index=True)
    name: Mapped[str] = mapped_column(nullable=False)
    organization_id: Mapped[uuid.UUID] = mapped_column(
        ForeignKey("organization.id"), nullable=False, index=True
    )

    @declared_attr
    def organization(cls) -> Mapped["Organization"]:
        return relationship("Organization", back_populates="message_groups")

    @declared_attr
    def messages(cls) -> Mapped[List["Message"]]:
        return relationship(back_populates="message_group")


class MessageGroupSync(CustomSyncBase, MessageGroupMixin):
    pass


class MessageGroup(CustomBase, MessageGroupMixin):
    pass



"""Redis track buckets"""

from pydantic import BaseModel, Field, ConfigDict
from typing import Annotated
from datetime import datetime, timezone
import uuid


TrackId = Annotated[uuid.UUID, Field(description="Track ID (UUID)")]

PlaylistId = Annotated[int, Field(description="Playlist ID (integer)")]

Timestamp = Annotated[int, Field(description="Unix timestamp")]


class RedisTrackBucketItem(BaseModel):
    """
    Represents a composite Redis key along with a timestamp.
    The Redis entry will be:
    { '{"track_id": "...", "playlist_id": ...}': ts }
    ts is the timestamp of utc now.
    """

    track_id: TrackId
    playlist_id: PlaylistId
    ts: Timestamp = Field(default_factory=lambda: int(datetime.now(timezone.utc).timestamp()))

    model_config = ConfigDict(json_encoders={uuid.UUID: str}, populate_by_name=True)

    def as_redis_entry(self) -> dict[str, int]:
        """
        Generate a Redis-ready dictionary:
        { '{"track_id": "...", "playlist_id": ...}': ts }
        """
        key = self.model_dump_json(include={"track_id", "playlist_id"})
        return {key: self.ts}

    @classmethod
    def from_redis_entry(cls, entry: dict[str, int]) -> "RedisTrackBucketItem":
        """
        Parse a Redis entry into a RedisTrackBucketItem.
        """
        return cls.model_validate_json(list(entry.keys())[0])


class RedisTrackBucketItemValue(BaseModel):
    """JSON string format for Redis keys: '{"track_id": "...", "playlist_id": ...}'"""

    track_id: TrackId
    playlist_id: PlaylistId



"""SQS messages."""

from datetime import datetime, timezone
from enum import Enum
from typing import Literal
from uuid import UUID


from artemis_model.redis.zone_state import SessionId
from artemis_model.schedule import ScheduleItem
from pydantic import BaseModel, Field

from artemis_model.zone_activity import ZoneActivityType


class Action(str, Enum):
    """Message type enum."""

    CALCULATE_ALL_SCHEDULES = "calculate-all-schedules"
    TRIAGE_SCHEDULE = "triage-schedule"
    TRIGGER_DATABASE_MIGRATION = "trigger-database-migration"
    MOVE_TIME_SLOT = "move-time-slot"
    PUSH_PLAYLIST = "push-playlist"
    EXPIRE_PUSHED_PLAYLIST = "expire-pushed-playlist"
    REFRESH_TRACK_BUCKET = "refresh-track-bucket"
    RECALCULATE_SCHEDULE = "recalculate-schedule"
    PLAYER_MODE_CHANGE = "player-mode-change"
    BAN_TRACK = "ban-track"


class BaseMessage(BaseModel):
    """Base message for incoming messages schema."""

    ts: str = Field(
        default_factory=lambda: datetime.now(timezone.utc).isoformat(),
        description="Timestamp value of the message",
    )
    action: Action


class MoveTimeSlotIn(BaseMessage):
    """Move to time slot message schema."""

    action: Action = Action.MOVE_TIME_SLOT
    zone_id: int
    timeslot: str


class TriageScheduleIn(BaseMessage):
    """Triage schedule message schema."""

    action: Action = Action.TRIAGE_SCHEDULE
    schedule: list[ScheduleItem]


class CalculateAllSchedulesIn(BaseMessage):
    """Calculate all schedules message schema."""

    action: Action = Action.CALCULATE_ALL_SCHEDULES


class PushPlaylistIn(BaseMessage):
    """Push playlist message schema."""

    action: Action = Action.PUSH_PLAYLIST
    zone_id: int
    playlist_ids: list[int]
    expire_at: datetime | None = Field(
        default=None, description="The datetime that the playlist will expire"
    )
    session_id: SessionId = Field(
        default_factory=lambda: int(datetime.now(timezone.utc).timestamp()),
    )


class PushPlaylistExpireIn(BaseMessage):
    """Push playlist expire message schema."""

    action: Action = Action.EXPIRE_PUSHED_PLAYLIST
    zone_id: int
    new_mode: Literal["pushplaylist", "scheduled"] = "scheduled"
    session_id: SessionId | None = None


class RefreshTrackBucketIn(BaseMessage):
    """Refresh track bucket message schema."""

    action: Action = Action.REFRESH_TRACK_BUCKET
    zone_id: int


class RecalculateScheduleIn(BaseMessage):
    """Recalculate schedule message schema."""

    action: Action = Action.RECALCULATE_SCHEDULE
    zone_id: int


class BanTrackIn(BaseMessage):
    """Ban track message schema."""

    action: Action = Action.BAN_TRACK
    zone_id: int
    track_id: UUID
    playlist_id: int


class ZoneActivityIn(BaseMessage):
    """Zone activity message schema."""

    action: Action = Action.ZONE_ACTIVITY
    zone_id: int
    activity_type: ZoneActivityType
    activity_data: dict



import uuid
from datetime import datetime
from typing import List

from sqlalchemy import UUID, Computed, ForeignKey, func, literal, text, ARRAY, Integer
from sqlalchemy.orm import Mapped, mapped_column, relationship

from artemis_model.base import CustomBase, CustomSyncBase, TSVector, TimeStampMixin

from sqlalchemy.ext.declarative import declared_attr


def to_tsvector_ix(*columns):
    s = " || ' ' || ".join(columns)
    return func.to_tsvector(literal("english"), text(s))


class TrackMixin(TimeStampMixin):
    id: Mapped[uuid.UUID] = mapped_column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    name: Mapped[str] = mapped_column(nullable=False)
    album_id: Mapped[uuid.UUID] = mapped_column(ForeignKey("album.id"), nullable=False)
    artist_id: Mapped[uuid.UUID] = mapped_column(ForeignKey("artist.id"), nullable=False)
    entry_date: Mapped[datetime] = mapped_column(nullable=False, default=datetime.utcnow)
    disabled: Mapped[bool] = mapped_column(nullable=False, default=False)
    legacy_id: Mapped[str] = mapped_column(nullable=False)
    is_internal: Mapped[bool] = mapped_column(nullable=False, default=False)
    is_parental_advisory: Mapped[bool] = mapped_column(nullable=False, default=False)
    decade: Mapped[int] = mapped_column(nullable=True)
    mood: Mapped[list[int]] = mapped_column(ARRAY(Integer), nullable=True)
    duration: Mapped[int] = mapped_column(nullable=True)

    name_tsv = mapped_column(
        TSVector(),
        Computed("to_tsvector('english', name)", persisted=True),
    )

    @declared_attr
    def track_playlist_associations(cls) -> Mapped[list["PlaylistTrackAssoc"]]:
        return relationship(back_populates="track", cascade="all, delete-orphan")

    @declared_attr
    def dj_set_track_associations(cls) -> Mapped[list["DjSetTrackAssoc"]]:
        return relationship(back_populates="track", cascade="all, delete-orphan")

    @declared_attr
    def album(cls) -> Mapped["Album"]:
        return relationship("Album", back_populates="tracks")

    @declared_attr
    def artist(cls) -> Mapped["Artist"]:
        return relationship("Artist", back_populates="tracks")

    @declared_attr
    def genres(cls) -> Mapped[List["Genre"]]:
        return relationship("Genre", secondary="track_genre_assoc", back_populates="tracks")

    @declared_attr
    def label(cls) -> Mapped["TrackLabel"]:
        return relationship(
            "TrackLabel",
            back_populates="track",
            cascade="all, delete-orphan",
        )

    # __table_args__ = (
    #     Index("fts_ix_track_name_tsv", to_tsvector_ix("name"), postgresql_using="gin"),
    # )


class TrackSync(CustomSyncBase, TrackMixin):
    pass


class Track(CustomBase, TrackMixin):
    pass


class TrackGenreAssocMixin:
    track_id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), ForeignKey("track.id"), primary_key=True
    )
    genre_id: Mapped[int] = mapped_column(ForeignKey("genre.id"), primary_key=True)


class TrackGenreAssocSync(CustomSyncBase, TrackGenreAssocMixin):
    pass


class TrackGenreAssoc(CustomBase, TrackGenreAssocMixin):
    pass


class TrackLabelMixin:
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    track_id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), ForeignKey("track.id"), primary_key=True
    )
    name: Mapped[str] = mapped_column(nullable=False)
    isrc: Mapped[str] = mapped_column(nullable=False)
    legacy_id: Mapped[str] = mapped_column(nullable=False)

    @declared_attr
    def track(cls) -> Mapped[Track]:
        return relationship("Track", back_populates="label")


class TrackLabelSync(CustomSyncBase, TrackLabelMixin):
    pass


class TrackLabel(CustomBase, TrackLabelMixin):
    pass
